;;; Version 0.03 of `font-lock-add-keywords' hackery.

;; From: Simon.Marshall@esrin.esa.it (Simon Marshall)

;; Put this file somewhere where Emacs can find it (i.e., in one of the paths
;; in your `load-path'), `byte-compile-file' it, and put in your ~/.emacs:
;;
;; (load "font-lock-add")
;;
;; assuming that the file is called `font-lock-add.el' of course.
;;
;; This is meant to be used by users; mode package writers should use
;; the appropriate mechanism to provide defaults etc. for their modes.

(provide 'font-lock-add)
(require 'font-lock)
(eval-when-compile
  (require 'cl))

;;;###autoload
(defun font-lock-add-defaults (major-mode defaults)
  "Add highlighting DEFAULTS for MAJOR-MODE.
MAJOR-MODE should be a symbol, the major mode symbol name, such as `c-mode'.
DEFAULTS should be a list; see the variable `font-lock-defaults'.  For example:

 (font-lock-add-defaults 'fubar-mode '(fubar-font-lock-keywords t))

adds one-level keywords-only highlighting defaults for Fubar mode."
  (push (cons major-mode defaults) font-lock-defaults-alist))

;;;###autoload
(defun font-lock-add-keywords (major-mode keywords &optional append level)
  "Add highlighting KEYWORDS for MAJOR-MODE.
MAJOR-MODE should be a symbol, the major mode command name, such as `c-mode'.
KEYWORDS should be a list; see the variable `font-lock-keywords'.

KEYWORDS are added at the beginning of the relevant MAJOR-MODE highlighting
list unless the optional third argument APPEND is non-nil, in which case
KEYWORDS are added at the end.

The optional fourth argument LEVEL, if non-nil, means add KEYWORDS to the
relevant MAJOR-MODE highlighting list for fontification level LEVEL.
Otherwise, add KEYWORDS based on the MAJOR-MODE fontification level specified
by the variable `font-lock-maximum-decoration'.  For example:

 (setq font-lock-maximum-decoration 2)
 (font-lock-add-keywords 'c-mode
  '((\"\\\\\\=<\\\\(FIXME\\\\):\" 1 font-lock-reference-face t)
    (\"\\\\\\=<\\\\(and\\\\|or\\\\|not\\\\)\\\\\\=>\" . font-lock-keyword-face)
    (\"\\\\\\=<FILE\\\\\\=>\"
     (0 font-lock-type-face)
     (font-lock-match-c++-style-declaration-item-and-skip-to-next nil nil
      (1 (if (match-beginning 4)
	     font-lock-function-name-face
	   font-lock-variable-name-face))))))

adds those three fontification patterns for C mode level 2 highlighting.
\(In order they mean (1) fontify `FIXME' words, even in comments, (2) fontify
`and', `or' and `not' words as keywords, and (3) fontify `FILE' words as types,
together with their following variable or function declaration items.)

If KEYWORDS is nil, means delete the relevant highlighting list.  Therefore, to
start a highlighting list from scratch, you can use a nil value first."
  (cond ((not (commandp major-mode))
	 ;; There is no command to turn the major mode on, but keep it simple
	 ;; and blame the luser.
	 (error "No `%s' mode to add keywords to" major-mode))
	((assq major-mode font-lock-defaults-alist)
	 ;; The variable name for the mode's relevant highlighting list is
	 ;; specified in `font-lock-defaults-alist'.  Find the name and set it.
	 (let ((variable (font-lock-choose-keywords
			  (nth 1 (assq major-mode font-lock-defaults-alist))
			  (font-lock-value-in-major-mode
			   (or level font-lock-maximum-decoration)))))
	   (if (null variable)
	       (error "No `%s' keywords to add to" major-mode)
	     (set variable
	      (cond ((null keywords) nil)
		    (append (append (symbol-value variable) keywords))
		    (t (append keywords (symbol-value variable))))))))
	((eq (car-safe (symbol-function major-mode)) 'autoload)
	 ;; The mode command is autoloaded, so we don't know what variable to
	 ;; set.  Arrange ourselves to be evaluated when its package is loaded.
	 (eval-after-load (nth 1 (symbol-function major-mode))
	  `(font-lock-add-keywords ',major-mode ',keywords ',append
	    (or ',level ',font-lock-maximum-decoration))))
	(t
	 ;; The variable name for the mode's relevant highlighting list is
	 ;; specified in `font-lock-defaults', but the mode command must be run
	 ;; to set its buffer local value.  Run it, find the name and set it.
	 (let ((buf-mode major-mode) (buf (get-buffer-create " *font-lock*")))
	   (unwind-protect
	       (save-excursion
		 (set-buffer buf)
		 (erase-buffer)
		 (funcall buf-mode)	; Uh-uh.  This could run hooks galore.
		 (let ((variable (font-lock-choose-keywords
				  (car font-lock-defaults)
				  (font-lock-value-in-major-mode
				   (or level font-lock-maximum-decoration)))))
		   (if (null variable)
		       (error "No `%s' keywords to add to" major-mode)
		     (set variable
		      (cond ((null keywords) nil)
			    (append (append (symbol-value variable) keywords))
			    (t (append keywords (symbol-value variable))))))))
	     (kill-buffer buf))))))

;;; end of `font-lock-add-keywords' hackery.

