; literate-mode.el
; mode for editing literate files - changes major-mode when a return is pressed 
; minor modes other than auto-fill will be turned off when the mode changes
; also gives some minor outlining functions (see below)

; This file is part of the LPAC Literate Programming System
; Copyright (c) 1993, 1994 London Parallel Application Centre
; 4 November 1993 Andrew Smallbone (andrew@dcs.qmw.ac.uk)
; modifications 25 April 1994 Andrew Smallbone (A.Smallbone@qmw.ac.uk)
;(define-key c-mode-map [return] 'indent-new-comment-line)


(setq literate-mode nil)

(defvar default-literate-language '"c"
  "The default language of code sections in literate-mode if no language option
is specified with a `\\begin{code}[lang]' or `\\setcodelang{lang}' statement" )

(defvar literate-section-search-string '"\\|\\\\\\(sub\\)*section"
  "This variable is used to allow a specified number of subsections to be
visible when in `outline' mode - broken at present")

(or (assq 'literate-mode minor-mode-alist)
    (setq minor-mode-alist (cons '(literate-mode " Literate") minor-mode-alist)))


(defun literate-mode (&optional arg dont-call-change-mode)
  "Toggle literate mode.
If first arg non nil turn on if +ve, off otherwise.
If second arg non nil don't change the current major mode."
  (interactive "P")
  (make-local-variable 'literate-mode)
  (setq literate-mode
        (if (null arg) (not literate-mode)
          (> (prefix-numeric-value arg) 0)))
  (if (and literate-mode (null dont-call-change-mode))
      (progn 
        (make-variable-buffer-local 'literate-current-mode)
        (setq literate-current-mode '"a")
        (literate-change-mode)
        (run-hooks 'literate-mode-hook))))

(defun literate-return ()
  "call literate-change-mode and insert a newline.
It is intended that the return key should be bound to this function 
so that literate-mode changes to the correct Major mode at 
reasonable periods."
  (interactive)
  (literate-change-mode)
  (newline 1))

(defun literate-change-mode ()
  "changes major mode appropriate for the current section of the file."
  (interactive)
  (if literate-mode
      (let ((was-fill-on auto-fill-function)
	    (mode (current-literate-mode)))
	(if (not (equal mode literate-current-mode))
	    (progn
	      (cond ((or (equal mode '"c") (equal mode '"m") (equal mode '"h")) (c-mode))
		    ((or (equal mode '"for") (equal mode '"FOR") (equal mode '"f") 
			 (equal mode '"F")) (fortran-mode))
		    ((equal mode '"f90") (f90-mode))
		    ((equal mode '"prl") (perl-mode))
		    ((equal mode '"latex") (latex-mode))
		    ((equal mode '"tcl") (tcl-mode))
		    (t (text-mode)
		       (message "unknown language `%s' using text-mode" mode)))
	      (setq literate-current-mode mode)
	      (if was-fill-on (turn-on-auto-fill))
	      (local-set-key [return] 'literate-return)
	      (set-buffer-modified-p (buffer-modified-p))
	      (setq selective-display t)
	      (literate-mode 1 1))))))

;; change this to only search back as far as needed
(defun current-literate-mode ()
  "Returns the language specifier of current section of file.
ie. if in code section return language, if documentation section
return latex etc."
  (interactive)
  (let ((first-point 0)
	(matched-data 0)
	(old-point (point-marker))
	(lang '"latex"))
    (save-excursion
      (if (re-search-backward "\\\\end\\{code\\}\\|\\\\begin\\{code\\}\\(\\[\\w+\\]\\)?" 0 t)
	  (if (char-equal (char-after (+ 1 (car (setq matched-data (match-data))))) ?b)               ;; if begin
	      (if (setq first-point (nth 2 matched-data))
		  (setq lang (buffer-substring  (+ 1 first-point) (-  (nth 3 matched-data) 1)))
		(if (re-search-backward "\\\\setcodelang\\{\\(\\w*\\)\\}" 0 t)        ;; look for default language 
		    (setq lang (buffer-substring  (nth 2 (setq matched-data (match-data))) (nth 3 matched-data)))
		  (setq lang default-literate-language)))))
      (goto-char old-point)
      lang)))


;;; now the hide code/section stuff
(defun literate-set-section-levels-visible (arg)
  "Set the number of subsections visible using literate hide mode"
  (interactive "nNumber of sub sections to display: ")
  (let ((sub '"")
	(num arg))
    (while (> num 0)
      (setq sub (concat sub "\\(sub\\)?"))
      (setq num (1- num)))
    (setq literate-section-search-string 
	  (concat "\\|\\\\" sub "section"))
    (if (<= arg 0) 
	(message "no subsections will be shown")
      (message "subsections will be displayed to %d levels" arg))
    (if literate-all-hidden (literate-show-all)
      (literate-hide-all-sections)
      (beginning-of-buffer))))


(defun literate-hide-this ()
  "hides the literate code section if the pointer is between a
\\begin{code} and \\end{code} or the entire section otherwise"
  (interactive)
  (let (was-buffer-modified (buffer-modified-p))
    (save-excursion
      (re-search-backward 
       (concat "\\\\begin\\{code\\}\\|\\\\end\\{code\\}\\|\\\\appendix\\|\\\\documentstyle" literate-section-search-string) nil t)
      (if (char-equal (char-after (+ (car (match-data)) 1)) ?b)
	  (literate-hide-this-code)
	(progn 
	  (goto-char (nth 1 (match-data)))
	  (literate-hide-this-section)))
      (set-buffer-modified-p was-buffer-modified))))

(defun literate-hide-this-code ()
  "hides the literate code section the cursor is located in
\(hides what's between the next \\end{code} and its preceeding \\begin{code}\)"
  (interactive)
  (let (was-buffer-modified (buffer-modified-p))
    (save-excursion
      (search-forward "\\end\{code\}")
      (literate-show/hide-code (- (point-marker) 11) ?\^M))
    (set-buffer-modified-p was-buffer-modified)))


(defun literate-hide-this-section ()
  "Hide current document section.
hides everything between next `\\section\{' or end of file and previous
`\\section{' or '\\documentstyle'"
  (interactive)
  (let (was-buffer-modified (buffer-modified-p))
  (save-excursion
    (if (re-search-forward (concat "\\\\appendix" literate-section-search-string) nil 1)
	(goto-char (- (car (match-data)) 1))
      (end-of-buffer))
    (setq to (point-marker))
    (if (re-search-backward 
	 (concat "\\\\appendix\\|\\\\documentstyle" literate-section-search-string) nil 1)
	(subst-char-in-region (point-marker) to ?\n ?\^M t)))
  (set-buffer-modified-p was-buffer-modified)))


(defun literate-show-this-code ()
  "shows the literate code section the cursor is located in"
  (interactive)
  (let (was-buffer-modified (buffer-modified-p))
    (save-excursion
      (search-forward "\\end\{code\}")
      (literate-show/hide-code (point-marker) ?\n))
    (set-buffer-modified-p was-buffer-modified)))
  
(defun literate-show-this-section ()
  "shows the previously hidden section the pointer is in.
Pointer should be between `section' commands"
  (interactive)
  (let (was-buffer-modified (buffer-modified-p))
    (save-excursion
      (if (re-search-forward 
	   (concat "\\\\appendix"  literate-section-search-string) nil 1)
	  (goto-char (- (car (match-data)) 1)))
      (setq to (point-marker))
      (if (re-search-backward 
	   (concat "\\\\appendix\\|\\\\\\(sub\\)*section\\|\\\\documentstyle" 
		   literate-section-search-string) nil 1)
	  (subst-char-in-region (point-marker) to ?\^M ?\n t)))
    (set-buffer-modified-p was-buffer-modified)))


(defun literate-show/hide-code (to flag)
  "hides or shows lines from `to' back to next \\begin{code}, according to flag.
If flag is `\\n' (newline character) then text is shown,
while if flag is `\\^M' (control-M) the text is hidden."
  (save-excursion
    (search-backward "\\begin\{code\}" nil t)
    (subst-char-in-region (point-marker) to (if (= flag ?\n) ?\^M ?\n) flag t)))

(defun literate-hide-all-code ()
  "hides all code sections in the literate file"
  (interactive)
  (let (was-buffer-modified (buffer-modified-p))
    (save-excursion
      (beginning-of-buffer)
      (while (search-forward "\\end\{code\}" nil t)
	(progn (literate-show/hide-code (- (point-marker) 11) ?\^M)
	       (next-line 1))))
    (set-buffer-modified-p was-buffer-modified)))


(defun literate-hide-all-sections ()
"hides everything except `section' and `documentstyle' commands and
the preamble before the documentstyle \(usally contains note and RCS etc.\)"
  (interactive)
  (let (was-buffer-modified (buffer-modified-p))
    (save-excursion
      (end-of-buffer)
      (setq to (point-marker))
      (while (re-search-backward 
	      (concat "\\\\appendix\\|\\\\documentstyle"  
		      literate-section-search-string) nil t)
	(progn
	  (setq next-to (- (car (match-data)) 1))
	  (subst-char-in-region (point-marker) to ?\n ?\^M t)
	  (setq to next-to))))
    (setq literate-all-hidden t)
    (set-buffer-modified-p was-buffer-modified)))


(defun literate-show-all ()
  "shows all previosly hidden text in the literate file"
  (interactive)
  (let (was-buffer-modified (buffer-modified-p))
    (save-excursion
      (end-of-buffer)
      (setq from (point-marker))
      (beginning-of-buffer)
      (subst-char-in-region from (point-marker)?\^M ?\n t))
    (setq literate-all-hidden nil)
    (set-buffer-modified-p was-buffer-modified)))

