/*   Author: Rob Bisseling.  Last modified: 15-8-1996.  Copyright reserved.*/

#include "bsplu.h"

int f(int k, int u, int R){
    /* Compute min(m: m*R+u>=k) for k>=0 and 0<=u<R */
    return ( u<k%R ?  k/R+1 : k/R ); 
}

void bsplu(int s, int t, int n, int *pi, double **a){
    /* Compute LU decomposition of A with partial pivoting 
       Program text for P(s,t) = processor s+t*M, with 0<=s<M and 0<=t<N 
       A is distributed according to the M by N grid distribution.
       Author: Rob Bisseling.  Last modified: 15-8-1996.  Copyright reserved.
    */

    double *vecallocd(int n);
    int *vecalloci(int n);
    int f(int k, int u, int R);
    double *ak, *ar, *uk, *lk, *gmax;
    int nr, nc, k, i, r, *gimax;

    nr=  f(n,s,M); /* number of local rows */
    nc=  f(n,t,N); /* same for columns */
    ak= vecallocd(nc);
    ar= vecallocd(nc);
    uk= vecallocd(nc);
    lk= vecallocd(nr);
    gmax= vecallocd(M);
    gimax= vecalloci(M);

    /* Initialise permutation vector pi */
    if (t==0){
        for(i=0; i<nr; i++){
            pi[i]= i*M+s; /* global row index */
        }
    }
    bsp_pushregister(pi,INT_MAX);
    bsp_pushregister(ak,INT_MAX);
    bsp_pushregister(ar,INT_MAX);
    bsp_pushregister(uk,INT_MAX);
    bsp_pushregister(lk,INT_MAX);
    bsp_pushregister(gmax,INT_MAX);
    bsp_pushregister(gimax,INT_MAX);
    bsp_pushregister(&r,INT_MAX);
    bsp_sync();

    for (k=0; k<n; k++){
        int kr, kc, kr1, kc1, j, imax, s1, t1;
        double absmax, max, pivot;
        kr= f(k,s,M); /* first local row with global index >= k */ 
        kr1= f(k+1,s,M); 
        kc= f(k,t,N); 
        kc1= f(k+1,t,N);
        
        if (k%N==t){   /* k=kc*N+t */
            /* Search for local absolute maximum in column k of A */
            absmax= 0.0;
            for (i=kr; i<nr; i++){
                if (fabs(a[i][kc])>absmax){
                    absmax= fabs(a[i][kc]);
                    imax= i;
                }
            } 
            if (absmax>EPS){
                max= a[imax][kc];
            } else {
                max= 0.0; imax= -1;
            }
        }
    
        if (k%N==t){
            /* Broadcast value and local index of maximum to P(*,t) */
            for(s1=0; s1<M; s1++){
                bsp_put(s1+t*M,&max,gmax,s*SZDBL,SZDBL);
                bsp_put(s1+t*M,&imax,gimax,s*SZINT,SZINT);
            }
        }
        bsp_sync();
        
        if (k%N==t){
            /* Determine global absolute maximum */
            absmax= 0.0; r=-1;
            for(s1=0; s1<M; s1++){
                if (fabs(gmax[s1])>absmax){
                    absmax= fabs(gmax[s1]);
                    r= gimax[s1]*M+s1; /* global index of pivot row */
                    pivot= gmax[s1];
                }
            }

        if (r==-1){ /* no pivot found */
            bsp_abort("bsplu at stage %d: Matrix is singular\n",k);
        }
            /* Broadcast index of pivot row to P(*,*) */
            for(t1=0; t1<N; t1++){
                bsp_put(s+t1*M,&r,&r,0,SZINT);
            }
        }
        bsp_sync();

        if (k%M==s){
            /* Store pi(k) in pi(r) on P(r%M,0) 
               No intermediate memory needed */
            if (t==0)
                bsp_put(r%M,&pi[k/M],pi,(r/M)*SZINT,SZINT);
            /* Store row k in ak on P(r%M,t) */
            bsp_put(r%M+t*M,a[k/M],ak,0,nc*SZDBL);
        }
        if (r%M==s){
            if (t==0)
                bsp_put(k%M,&pi[r/M],pi,(k/M)*SZINT,SZINT);
            bsp_put(k%M+t*M,a[r/M],ar,0,nc*SZDBL);
        }
        bsp_sync();

        if (k%M==s){   /* k=kr*M+s and hence kr=k/M */
            /* Assign row r to A(k,*) */
            for(j=0; j<nc; j++){
                a[kr][j]= ar[j];
            }
        }
        if (r%M==s){
            for(j=0; j<nc; j++){
                a[r/M][j]= ak[j];
            }
        }
        if (k%M==s){ 
            /* Store new row k in uk - first phase */
            for(j=kc1; j<nc; j++){     
                bsp_put(j%M+t*M,&a[kr][j],uk,j*SZDBL,SZDBL);
            }
        }
        if (k%N==t){   /* k=kc*N+t */
            /* Compute column k and store it in lk - first phase */
            for(i=kr1; i<nr; i++){     
                a[i][kc] /= pivot;
                bsp_put(s+(i%N)*M,&a[i][kc],lk,i*SZDBL,SZDBL);
            }
        }
        bsp_sync();
       
        /* Store row k in uk - second phase */
        for(j=f(kc1,s,M)*M+s; j<nc; j +=M){
            for(s1=0; s1<M; s1++){
                bsp_put(s1+t*M,&uk[j],uk,j*SZDBL,SZDBL);
            }
        }
        /* Store column k in lk - second phase */
        for(i=f(kr1,t,N)*N+t; i<nr; i +=N){
            for(t1=0; t1<N; t1++){
                bsp_put(s+t1*M,&lk[i],lk,i*SZDBL,SZDBL);
            }
        }
        bsp_sync();
      
        /* Rank-1 update of A */
        for(i=kr1; i<nr; i++){
            for(j=kc1; j<nc; j++){
                a[i][j] -= lk[i]*uk[j];
            }
        }
    }
    bsp_popregister(&r);
    bsp_popregister(gimax);
    bsp_popregister(gmax);
    bsp_popregister(lk);
    bsp_popregister(uk);
    bsp_popregister(ar);
    bsp_popregister(ak);
    bsp_popregister(pi);
    bsp_sync();
} /* end bsplu */
