/*
   This is a test program which uses bsplu to decompose a matrix A into 
   triangular factors L and U, with partial row pivoting.
   The decomposition is A(pi(i),j)=(LU)(i,j), all i,j,
   where pi is a permutation.
 
   The input matrix A is defined by: B(i,j)= 0.5*i+1     if i<=j 
                                             0.5*j+0.5   if i>j,
                                     A(i,j)=B((i-1) mod n, j).
   The indices i,j have a range: 0 <= i,j < n.
   The output matrix L is defined by: L(i,j)=0   if i<j,
                                            =1   if i=j,
                                            =0.5 if i>j.
   The output matrix U is defined by: U(i,j)=1   if i<=j,
                                            =0   if i>j.
   The permutation pi is defined by:  pi(i)= (i+1) mod n.

   The matrices L and U overwrite A.
   Output of L and U is in triples (i,j,lu(i,j)):
     (i,j,0.5) for i>j
     (i,j,1) for i<=j
   Output of pi is in pairs (i,pi(i))
     (i,(i+1) mod n) for all i.
   For verification of correct execution:
   check the triples of L/U and the pairs of pi.

   The matrices are chosen such that there is a unique choice of pivots.
   In phase k of the LU decomposition, row k is swapped with row r=k+1.
   For the M x N grid distribution this forces a row swap
   between processor rows. 
                                        
   Author: Rob Bisseling.  Last modified: 15-8-1996.  Copyright reserved.
*/
#include "bsplu.h"

int M,N;


void bsplu_test(){
    int *vecalloci(int n);
    double **matallocd(int m, int n);
    int f(int k, int u, int R);
    void bsplu(int s, int t, int n, int *pi, double **a);
    int p, pid, q, s, t, n, nr, nc, i, j, iglob, jglob, *pi;
    double **a;
  
    bsp_begin(M*N);
    p=bsp_nprocs(); /* p=M*N */
    pid=bsp_pid();

    bsp_pushregister(&M,INT_MAX);
    bsp_pushregister(&N,INT_MAX);
    bsp_pushregister(&n,INT_MAX);
    bsp_sync();

    if (pid==0){
        printf("Please enter matrix size n:\n");
        scanf("%d",&n);
        for (q=0; q<p; q++){
            bsp_put(q,&M,&M,0,SZINT);
            bsp_put(q,&N,&N,0,SZINT);
            bsp_put(q,&n,&n,0,SZINT);
        }
    }
    bsp_sync();

    s= pid%M;  /* 0 <= s < M */
    t= pid/M;  /* 0 <= t < N */

    /* Allocate and initialise matrix */
    nr=  f(n,s,M); /* number of local rows */
    nc=  f(n,t,N); /* same for columns */
    pi= vecalloci(nr);
    a= matallocd(nr,nc);
  
    if (pid==0){
        printf("LU decomposition of %d by %d matrix\n",n,n);
        printf("using %d times %d processors\n",M,N);
    }
    for (i=0; i<nr; i++){
        iglob= i*M+s; /* Global row index in A */
        iglob= (iglob-1+n)%n; /* Global row index in B */
        for (j=0; j<nc; j++){
            jglob= j*N+t; /* Global column index */
            a[i][j]= (iglob<=jglob ? 0.5*iglob+1 : 0.5*(jglob+1) );
        }
    }
    bsp_sync();
  
    if (pid==0){
        printf("Start of LU decomposition\n");
    }
    bsplu(s,t,n,pi,a);
    if (pid==0){
        printf("End of successful LU decomposition\n");
    }
    bsp_sync();

    if (t==0){
        for (i=0; i<nr; i++){
            iglob=i*M+s;
            printf("i=%d, pi=%d, proc=(%d,%d)\n",iglob,pi[i],s,t);
        }
    }
    bsp_sync();

    for (i=0; i<nr; i++){
        iglob=i*M+s;
    for (j=0; j<nc; j++){
        jglob=j*N+t;
        printf("i=%d, j=%d, a=%lf, proc=(%d,%d)\n",iglob,jglob,a[i][j],s,t);
        }
    }
    bsp_popregister(&n);
    bsp_popregister(&N);
    bsp_popregister(&M);
    bsp_sync();

    bsp_end();
}

int main(int argc, char **argv){
 
    bsp_init(bsplu_test, argc, argv);
    printf("Please enter number of processor rows M:\n");
    scanf("%d",&M);
    printf("Please enter number of processor columns N:\n");
    scanf("%d",&N);
    if (M*N > bsp_nprocs()){
        printf("Not enough processors available: %d wanted, %d available\n",
                M*N, bsp_nprocs());
        exit(1);
    }
    bsplu_test();
    exit(0);
}

