
/* matmult.c */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <time.h>

#include "bsp.h"

#include <sys/types.h>
#include <sys/time.h>
#include <assert.h>
#include "matmult.h"

/*************************/
/* Matrix Multiplication */
/*************************/

void main(void)
{
  globals glo;
  globals *g;
  int     i;
  int     dest;

  bsp_begin(bsp_nprocs());

  /*****************************/
  /* set some global variables */
  /*****************************/

  g = &glo;
  g->myProc = bsp_pid(); 
  g->numProcs = bsp_nprocs();
  for (g->rootProcs = 1; g->rootProcs * g->rootProcs < g->numProcs;
       g->rootProcs++);
  if (g->rootProcs * g->rootProcs > g->numProcs)
    error("Number of processors not a square!", g);

  /***************************/
  /* read the input matrices */
  /***************************/

  getInput(g);

  /* and set result matrix to 0 */
  for (i = 0; i < g->locMatSizeSq; i++)
    g->matC[i] = 0.0;

  /************************************************/
  /* multiply the matrix using Cannon's algorithm */ 
  /************************************************/

  seqMatMult(g->matA, g->matB, g->matC, g);

  /* the other local multiplies between newly shifted-in blocks */
  for (i = 1; i < g->rootProcs; i++)
  {
    /* shift first matrix one processor to the right */
    dest = g->myProc + 1;
    if (dest % g->rootProcs == 0) 
      dest -= g->rootProcs;
    sendMat(g->matA, dest, g);

    /* shift second matrix one processor down */
    dest = (g->myProc + g->rootProcs) % g->numProcs;
    sendMat(g->matB, dest, g);

    /* local multiply between shifted local blocks */
    seqMatMult(g->matA, g->matB, g->matC, g);
  }

  /************************/
  /* output result matrix */
  /************************/

  finalOutput(g);

  bsp_end();

  return;
}



/*****************/
/* Get the input */
/*****************/

void getInput(globals *g)
{
  FILE   *infile;
  int    i;
  int    row;
  int    dest;
  int    destRow;
  int    stat;
  double *index;
  int    logical_row;
  int    logical_column;
  int    physical_row;
  int    physical_column;

  /* open file */
  if (g->myProc == 0)
  {
    if ((infile = fopen(INFILE, "r")) == NULL)
      error("Input file could not be opened!\n", g);

    /* read size of input matrices */
    fscanf(infile, "%d\n", &(g->matSize));
    for (i = 0; i < g->numProcs; i++)
      bsp_send(i, NULL, (void *) &(g->matSize), sizeof(int));
  }
  
  /* receive and store matrix size */
  bsp_sync();
  bsp_move(&(g->matSize), sizeof(int));


  /* now compute size of local parts of matrix */
  if (g->matSize % g->rootProcs != 0)
    error("Matrix size not a multiple of root of number of processors!", g);
  g->locMatSize = g->matSize / g->rootProcs;
  g->locMatSizeSq = g->locMatSize * g->locMatSize;
  g->locFirstRow = (g->myProc / g->rootProcs) * (g->matSize / g->rootProcs);
  g->locFirstCol = (g->myProc % g->rootProcs) * (g->matSize / g->rootProcs);

  /* and allocate memory for local matrices */
  g->matA = (double *) calloc(g->locMatSize * g->locMatSize, sizeof(double));
  g->matB = (double *) calloc(g->locMatSize * g->locMatSize, sizeof(double));
  g->matC = (double *) calloc(g->locMatSize * g->locMatSize, sizeof(double));
  if (g->matA == NULL || g->matB == NULL || g->matC == NULL)
    error("Not enough memory to store local matrices!", g);

  /* read in both matrices */
  /* Remember initial conditions.
     Block A[i,j] -> P[i,(j-i)mod(sqrt(P)]
     Block B[i,j] -> P[(i-j)mod(sqrt(P),j] */

  for (i = 0; i < 2; i++)
    for (row = 0; row < g->matSize; row++)
      {
	
	/* have proc 0 read in one row of the matrix, and send it out */
	
	if (g->myProc == 0)
	  {
	    /* Proc 0 read in one row of the matrix. */
	    /* Use g->matC for temporary storage. */
	    
	    /* starting dest proc */
	    /* dest = (row / g->locMatSize) * g->rootProcs; */

	    /* row number is enough to select the logical processor row */
	    logical_row = row / g->locMatSize;
	    
	    for (logical_column = 0; logical_column < g->rootProcs;
		 logical_column++)
	      {
		/* Do mapping from logical indices to physical indices to
		   process number. */

		if (i == 0) {
		  physical_row = logical_row;
		  physical_column = (logical_column - logical_row +
				     g->rootProcs) % g->rootProcs;
		}
		else {
		  physical_row = (logical_row - logical_column +
				  g->rootProcs) % g->rootProcs;
		  physical_column = logical_column;
		}

		dest = physical_row * g->rootProcs + physical_column;

		for (index = g->matC; index < g->matC + g->locMatSize; index++)
		  fscanf(infile, "%lf\n", index);
		bsp_send(dest, NULL, (void *) g->matC,
			 g->locMatSize * sizeof(double));
	      }
	  }
	
	/* sync, and put received matrix elements into local matrix */
	bsp_sync();

	
	bsp_get_tag(&stat, NULL);

	if (stat != -1)
	  {
	    destRow = row % g->locMatSize;
	    if (i == 0)
	      index = g->matA + (destRow * g->locMatSize);
	    else
	      index = g->matB + (destRow * g->locMatSize);
	    bsp_move((void *) index, g->locMatSize * sizeof(double));
	  }
      }
 
  /* Proc 0 closes the file. */
  if (g->myProc == 0)
    fclose(infile);

  return;
}



/*************************************/
/* send matrix to a given processor  */
/*************************************/

void sendMat(double *mat, int dest, globals *g)
{
  bsp_send(dest, NULL, mat, g->locMatSizeSq * sizeof(double));

  bsp_sync();

  bsp_move(mat, g->locMatSizeSq * sizeof(double));

  return;
}



/*****************************************************/
/* Sequential blocked multiplication of two matrices */
/*****************************************************/

void seqMatMult(double *A, double *B, double *C, globals *g)
{
  double A0[BLOCKSIZE][BLOCKSIZE];
  double B0[BLOCKSIZE][BLOCKSIZE];
  int i, j, k;
  int i0, j0, k0;
  int j1, k1;
  int imax, jmax, kmax;
  int lms = g->locMatSize;
  double x;

  for (i = 0; i < lms; i += BLOCKSIZE)
  {
    /* set upper bound for i0 index */
    if ((imax = lms - i) > BLOCKSIZE)
      imax = BLOCKSIZE;

    for (j = 0; j < lms; j += BLOCKSIZE)
    {
      /* set upper bound for j0 index */
      if ((jmax = lms - j) > BLOCKSIZE)
        jmax = BLOCKSIZE;

      /* copy from matrix into block */
      for (i0 = 0; i0 < imax; i0++)
        for (j0 = 0, j1 = (i + i0) * lms + j; j0 < jmax; j0++)
          A0[i0][j0] = A[j1+j0];

      for (k = 0; k < lms; k += BLOCKSIZE)
      {
        /* set upper bound for k0 index */
        if ((kmax = lms - k) > BLOCKSIZE)
          kmax = BLOCKSIZE;

        /* copy from matrix to other block */
        for (j0 = 0; j0 < jmax; j0++)
          for (k0 = 0, k1 = (j + j0) * lms + k; k0 < kmax; k0++)
            B0[j0][k0] = B[k1+k0];        

        /* multiply the two blocks */
        for (i0 = imax; i0-- > 0;)
          for (k0 = kmax, k1 = (i + i0) * lms + k; k0-- > 0;)
          {
            x = 0.0;
            for (j0 = jmax; j0-- > 0;)
              x += A0[i0][j0] * B0[j0][k0];
            C[k1+k0] += x;
          }
      }
    }
  }

  return;
}



/*******************************/
/* Write result matrix to file */
/*******************************/

void finalOutput(globals *g)
{
  FILE   *outfile;
  int    i;
  int    row;
  int    tag;
  int    dummyTag;
  int    status;
  int    tagSize;
  double x;
  double *outRow;
  double *index;

  /* This is the only section where we use a tag, since it is the */
  /* only function in which any process receives more than one message. */

  tagSize = sizeof(int);
  bsp_set_tag_size(&tagSize);

  tag = g->myProc % g->rootProcs; /* Index of the proc in its row. */

  /* Proc 0 opens file */
  if (g->myProc == 0)
  {
    if ((outfile = fopen(OUTFILE, "w")) == NULL)
      error("Output file could not be opened!\n", g);

    /* allocate space for output buffer for one row */
    outRow = (double *) calloc(g->matSize, sizeof(double));
  }

  /* output result matrix */

  for (row = 0; row < g->matSize; row++)
  {
    /* if the processor contains elements from the current row */
    if (row >= g->locFirstRow && row < g->locFirstRow + g->locMatSize)
      {
	index = g->matC + (row % g->locMatSize) * g->locMatSize;
	bsp_send(0, &tag, index, g->locMatSize * sizeof(double));
      }

    /* sync, and have proc 0 put received elements into buffer */
    bsp_sync();

    if (g->myProc == 0)
      {
	bsp_get_tag(&status, &dummyTag);
	while(status != -1)
	  {
	    index = outRow + dummyTag * g->locMatSize;
	    bsp_move(index, g->locMatSize * sizeof(double));
	    bsp_get_tag(&status, &dummyTag);
	  }

      /* and then into the file */
      for (i = 0; i < g->matSize; i++)
        fprintf(outfile, "%g\n", outRow[i]);
    }
  }

  /* close the file */
  if (g->myProc == 0)
    fclose(outfile);
}



/*********************************/
/* Print error message and abort */
/*********************************/

error(char *text, globals *g)
{
  printf("Proc #%d: ", g->myProc);
  printf(text);
  exit(0);
}

